////////////////////////////////////////////////////////////////////////////////
// HIDsample										Create Date	2015/11/12
//				for PIC18F14K50						Last Update	2018/04/06
// file name :  main.c										Saka Softwares
// language  :  XC8 (v1.45)
////////////////////////////////////////////////////////////////////////////////
// 新型2KB版 HID bootloadr 及び usbhiddev.c の動作サンプルです。
// 動作確認は、書き込み Microchips製 USB Bootloader v2.15
// 及び HIDConsole Version 1.5で行っています。
// このプログラムの目的は、C18言語からXC8言語に移行するために作りました。
// のでC18でのコンパイルは出来ません。
////////////////////////////////////////////////////////////////////////////////
// Include Files ///////////////////////////////////////////////////////////////
#include <xc.h>
#include <stdint.h>
#include <stdbool.h>

#include "usbhiddev.h"

// Version info ////////////////////////////////////////////////////////////////
//注. 下記 VersionWord は、XC8 (v1.36)で旨く書き込めない場合があります。
//    HID bootloadrを使わない場合コメントアウトしてください。
const uint16_t VersionWord @ 0x0816 = 0x0150;	// BCD表現のバージョン No.
#define	SOFTWEAR_VERSION	"sample_hid18f14k50 (XC8) Ver1.5"
////////////////////////////////////////////////////////////////////////////////
//	1.5		2018/04/06	usbhiddev.cをPIC16/PIC18 両方に対応
//	1.4		2018/02/07	xc8 Ver1.45時のワーニングに対応 LEDの点滅処理変更
//	1.2		2016/03/12	usbhiddevの不具合修正
//	1.0		2015/11/19	公開バージョン
////////////////////////////////////////////////////////////////////////////////

// Configuration ///////////////////////////////////////////////////////////////
// HIDブートローダーと同一コンフィグレーションにしてください。
// pic18f14k50_xc8_bootloader.hexに対応
#pragma config  CPUDIV	= NOCLKDIV	// CPU システムクロック選択ビット
#pragma config  USBDIV	= OFF		// USB クロック選択ビット
#pragma config  FOSC	= HS		// オシレータ選択ビット
#pragma config  PLLEN	= ON		// 4 X PLL イネーブルビット
#pragma config  PCLKEN	= OFF		// プライマリ クロック イネーブルビット
#pragma config  FCMEN	= OFF		// フェイルセーフ クロック モニタ イネーブルビット
#pragma config  IESO	= OFF		// 内部/外部オシレータ切り換えビット
#pragma config  PWRTEN	= OFF		// パワーアップ タイマ イネーブルビット (OFFで有効)
#pragma config  BOREN	= ON		// ブラウンアウト リセット イネーブルビット
#pragma config  BORV	= 30		// ブラウンアウト リセット電圧ビット
#pragma config  WDTEN	= OFF		// ウォッチドッグ タイマ イネーブルビット
#pragma config  WDTPS	= 32768		// ウォッチドッグ タイマ ポストスケーラ選択ビット
#pragma config  HFOFST	= OFF		// HFINTOSC 高速起動ビット
#pragma config  MCLRE	= OFF		// MCLR ピン イネーブルビット
#pragma config  STVREN	= ON		// スタックフル/ アンダーフロー リセット イネーブルビット
#pragma config  LVP		= OFF		// 単電源ICSP イネーブルビット
#pragma config  BBSIZ	= ON		// ブートブロック サイズ選択ビット(2kW boot block size)
#pragma config  XINST	= OFF		// 拡張命令セットイネーブルビット (ENHCPU)
#pragma config  CP0		= OFF		// コード保護ビット
#pragma config  CP1		= OFF		// コード保護ビット
#pragma config  CPB		= OFF		// ブートブロック コード保護ビット
#pragma config  CPD		= OFF		// EEPROM コード保護ビット
#pragma config  WRT0	= OFF		// 書き込み保護ビット
#pragma config  WRT1	= OFF		// 書き込み保護ビット
#pragma config  WRTC	= OFF		// コンフィグレーション レジスタ書き込み保護ビット
#pragma config  WRTB	= OFF		// ブートブロック書き込み保護ビット
#pragma config  WRTD	= OFF		// EEPROM 書き込み保護ビット
#pragma config  EBTR0	= OFF		// テーブル読み出し保護ビット
#pragma config  EBTR1	= OFF		// テーブル読み出し保護ビット
#pragma config  EBTRB	= OFF		// ブートブロック テーブル読み出し保護ビット

// Hardware Settings ///////////////////////////////////////////////////////////
// LEDs
#define LED_ON		1
#define LED_OFF		0
#define mLED1		LATBbits.LATB6
#define mLED1Tris	TRISBbits.TRISB6

// SWITCH
#define SW_ON		0
#define SW_OFF		1
#define sw2			PORTAbits.RA3

// Port入出力
#define PORTOUT		0
#define PORTIN		1

// Variable ////////////////////////////////////////////////////////////////////
#define DATA_SIZE	80			// EP1Outと同じか大きくしてください。
uint8_t rcv_dat[DATA_SIZE];		// 受信データ

union {
	uint8_t Val;				//フラグ値
	struct {
		unsigned VER:1;			//バージョン送信フラグ
		unsigned HELP:1;		//ヘルプ送信フラグ
		unsigned ECHO:1;		//エコーフラグ
		unsigned :6;
	}bits;
} FLAG;							// 各種フラグ

// Fixed number ////////////////////////////////////////////////////////////////
// ヘルプデータ
const uint8_t *HELP[] = {
	"Help",
	" v : Version",
	" e : Echoback ON",
	" E : Echoback OFF",
	" h : Help",
	" Boot : Bootloder mode",
	" ",
	""
};

// Internal Function ///////////////////////////////////////////////////////////
void main(void);
void interrupt low_priority IsrLow(void);
void ProcessIO(void);
void bootloader(uint8_t* pbuf);
void MainUsbStatus(void);

////////////////////////////////////////////////////////////////////////////////
// メイン処理部
// 引  数	: 無し
// 戻り値	: 無し
////////////////////////////////////////////////////////////////////////////////
void main(void)
{
	ClrWdt();
	INTCON  = 0x00;					// 割り込み禁止
	STKPTR  = 0x00;					// スタックポインタ設定
	WDTCON  = 0x01;					// ウォッチドックタイマ起動
	RCONbits.IPEN	= 1;			// 優先度レベルの設定を有効にす
	IPR2bits.USBIP	= 0;			// USB 割り込み優先度:Low
	FLAG.Val = 0x00;				// フラグクリア
#if defined(ENABLE_USB_LED_BLINK_STATUS)
	mLED1 = LED_OFF;
	mLED1Tris = PORTOUT;
#endif
	USBDeviceInit();				// 割り込み対応USBの初期化
	INTCONbits.GIEL = 1;			// 低割り込み許可
	INTCONbits.GIEH = 1;			// 高割り込み許可
	while (1) {
		ClrWdt();
		bootloader(rcv_dat);		// ブートローダー移行処理を行います。
		USBDeviceTasks();			// USBのバス接続切断処理を行います。
		MainUsbStatus();			// USBフラグを見てLEDの点灯処理を行います。
		ProcessIO();				// 通信処理
    }    
}    

////////////////////////////////////////////////////////////////////////////////
// USB割り込み処理部処理部
// 引  数	: 無し
// 戻り値	: 無し
// USBのプラグ&プレイ関連の処理を行っています。必須です。 
////////////////////////////////////////////////////////////////////////////////
void interrupt low_priority IsrLow(void)
{
	// USB割り込み処理
	USBInterruptTasks();
}

////////////////////////////////////////////////////////////////////////////////
// プロセスＩＯ処理部
// 引  数	: 無し
// 戻り値	: 無し
// USB関連の処理を一括で行っています。
// HID受信した場合フラグを立てて各処理を行い、HID送信が可能になるまで待ち
// 結果を返す仕様にしています。
// 1回の受信または送信は64byteです。HIDのリポートディスクリプタで設定しています。
// 送信処理(HidTxRamReport)は、前の送信が終わるまで待ちます。
// 複数送る場合ウォッチドックのタイムアウトに気を付けてください。
////////////////////////////////////////////////////////////////////////////////
void ProcessIO(void)
{
	uint8_t len;
	if ((len = HidRxReport(rcv_dat, DATA_SIZE))  < RCV_RECEING) {
		// エコーの送信 /////////////////////////////////////////////////////////
		if (!FLAG.bits.ECHO) HidTxRamReport(rcv_dat, len);
		switch (rcv_dat[0]) {
			// バージョン情報の送信 /////////////////////////////////////////////
			case 'v' : {
				HidTxRomReport(SOFTWEAR_VERSION, sizeof(SOFTWEAR_VERSION));
				break;
			}
			// エコーバックのON /////////////////////////////////////////////////
			case 'e' : {
				FLAG.bits.ECHO = 0;
				break;
			}
			// エコーバックのOFF ////////////////////////////////////////////////
			case 'E' : {
				FLAG.bits.ECHO = 1;
				break;
			}
			// ヘルプの送信 /////////////////////////////////////////////////////
			case 'h' : {
				uint8_t cnt = 0;
				while ( *HELP[cnt] != '\0') {
					HidTxRomReport(HELP[cnt++], 22);
				}
				break;
			}
			default	 : break;
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
// ブートローダー処理部
// 引  数	: 受信データバッファポインタ
// 戻り値	: 無し
// sw2または受信コマンドによりHIDブートローダーへ移行します。
// PIC 新型2KB版 HID Bootloaderに対応しています。
////////////////////////////////////////////////////////////////////////////////
void bootloader(uint8_t* pbuf)
{
	if (!sw2 || (pbuf != 0 && *(pbuf+0) == 'B'
		&& *(pbuf+1) == 'o' && *(pbuf+2) == 'o' && *(pbuf+3) == 't')) {
		INTCONbits.GIEL = 0;				// 低割り込み禁止
		INTCONbits.GIEH = 0;				// 高割り込み禁止
		LATA  = 0;							// ポートAクリア
		LATB  = 0;							// ポートBクリア
		LATC  = 0;							// ポートcクリア
		TRISA = 0;							// ポートA初期化
		TRISB = 0;							// ポートB初期化
		TRISC = 0;							// ポートC初期化
		uint16_t cnt = 0xFFFF;
		while (cnt--) { 
			ClrWdt();
			UCON = 0;						// ウェイト付きUSB停止
		}
		asm("goto 0x001C");					// bootloader Entryへ
	}
}

////////////////////////////////////////////////////////////////////////////////
// USBステータスによる各処理部
// 引  数	: 無し
// 戻り値	: 無し
// LED表示処理
// USBに未接続時は消灯 USB接続処理中は点灯 USBの通信中は点滅します。
//* サスペンド処理
// サスペンド時にスリープします。
//* その前後にユーザー処理を入れます。
////////////////////////////////////////////////////////////////////////////////
void MainUsbStatus(void)
{
#ifdef ENABLE_USB_LED_BLINK_STATUS
	// USB ステータス表示処理 ///////////////////////////////////////////////////
	static uint16_t led_count = 0;
	led_count--;
	if(led_count == 0) {
		led_count = 0x3000;
		if (USBGetDeviceState() == DETACHED_STATE) {
			mLED1 = LED_OFF;			// 非接続	(消灯)
		}
		else if (USBGetDeviceState() < CONFIGURED_STATE) {
			mLED1 = (unsigned)!mLED1;	// 接続中	(点滅)
		}
		else {
			mLED1 = LED_ON;				// 通信中	(点灯)
		}
	}
#endif
	// サスペンド処理 ///////////////////////////////////////////////////////////
	if (USBGetDeviceState() == CONFIGURED_STATE && USBIsSuspended() ) {
	// メインのサスペンド移行処理部
#ifdef ENABLE_USB_LED_BLINK_STATUS
		mLED1 = LED_OFF;				// 通信中断	(消灯)
#endif
		// サスペンド中何もせずスリープする
		while (!USBIsActive() && !RCONbits.TO) {
			Sleep();
		}
		// メインのウェイクアップ処理部
#ifdef ENABLE_USB_LED_BLINK_STATUS
		mLED1 = LED_ON;					// 通信開始 (点灯)
#endif
	}
}
// EOF main.c //////////////////////////////////////////////////////////////////
